import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';

import 'lyrics.dart';

class LyricPainter extends CustomPainter with ChangeNotifier {
  //Lyrics list
  List<Lyrics> lyrics;

  //Translation/Transliteration Lyrics List
  List<Lyrics> subLyrics;

  //Canvas size
  Size canvasSize = Size.zero;

  //Maximum font width
  double lyricMaxWidth;

  //Lyric spacing
  double lyricGapValue;

  //Lyric spacing
  double subLyricGapValue;

  //Control lyrics sliding by offset
  double _offset = 0;

  set offset(value) {
    _offset = value;
    notifyListeners();
  }

  get offset => _offset;

  //Lyric position
  int _currentLyricIndex = 0;

  get currentLyricIndex => _currentLyricIndex;

  set currentLyricIndex(value) {
    _currentLyricIndex = value;
    notifyListeners();
  }

  //Lyric line height
  double _subLyricHeight = 0;

  //Lyric style
  TextStyle lyricTextStyle;

  //Sliding lyrics style
  TextStyle draggingLyricTextStyle;

  //Sliding lyrics style
  TextStyle draggingSubLyricTextStyle;

  //Translation/transliteration lyrics style
  TextStyle subLyricTextStyle;

  //Current lyrics style
  TextStyle currLyricTextStyle;

  //Current translation/transliteration lyrics style
  TextStyle currSubLyricTextStyle;

  //Swipe to the row
  int _draggingLine;

  get draggingLine => _draggingLine;

  set draggingLine(value) {
    this._draggingLine = value;
    notifyListeners();
  }

  //Lyric brush array
  final List<TextPainter> lyricTextPaints;

  //Translation/transliteration lyrics brush array
  final List<TextPainter> subLyricTextPaints;

  LyricPainter(this.lyrics, this.lyricTextPaints, this.subLyricTextPaints,
      {this.subLyrics,
      double currDuration,
      TickerProvider vsync,
      this.lyricTextStyle,
      this.subLyricTextStyle,
      this.currLyricTextStyle,
      this.currSubLyricTextStyle,
      this.draggingLyricTextStyle,
      this.draggingSubLyricTextStyle,
      this.lyricGapValue,
      this.subLyricGapValue,
      this.lyricMaxWidth});

  @override
  void paint(Canvas canvas, Size size) {
    canvasSize = size;

    //The Y coordinate of the initial lyrics is in the center
    lyricTextPaints[currentLyricIndex]
      //Set lyrics
      ..text = TextSpan(
          text: lyrics[currentLyricIndex].lyrics, style: currLyricTextStyle)
      ..layout(maxWidth: lyricMaxWidth);
    var currentLyricY = _offset +
        size.height / 5 -
        lyricTextPaints[currentLyricIndex].height / 2;

    //Traverse the lyrics to draw
    for (int lyricIndex = 0; lyricIndex < lyrics.length; lyricIndex++) {
      var currentLyric = lyrics[lyricIndex];
      var isCurrLine = currentLyricIndex == lyricIndex;
      var isDraggingLine = _draggingLine == lyricIndex;
      var currentLyricTextPaint = lyricTextPaints[lyricIndex]
        //Set lyrics
        ..text = TextSpan(
            text: currentLyric.lyrics,
            style: isCurrLine
                ? currLyricTextStyle
                : isDraggingLine
                    ? draggingLyricTextStyle
                    : lyricTextStyle);
      currentLyricTextPaint.layout(maxWidth: lyricMaxWidth);
      var currentLyricHeight = currentLyricTextPaint.height;
      //Lyrics only drawn on the screen
      if (currentLyricY < size.height && currentLyricY > 0) {
        //Draw lyrics to canvas
        currentLyricTextPaint
          ..paint(
              canvas,
              Offset((size.width - currentLyricTextPaint.width) / 2,
                  currentLyricY));
      }
      //After the current lyrics are over, adjust the y coordinate of the next lyrics
      currentLyricY += currentLyricHeight + lyricGapValue;
      //If there are translated lyrics, look for the translated lyrics of the line in the future
      if (subLyrics != null) {
        List<Lyrics> remarkLyrics = subLyrics
            .where((subLyric) =>
                subLyric.startTime >= currentLyric.startTime &&
                subLyric.endTime <= currentLyric.endTime)
            .toList();
        remarkLyrics.forEach((remarkLyric) {
          //Get location
          var subIndex = subLyrics.indexOf(remarkLyric);

          var currentSubPaint = subLyricTextPaints[subIndex] //Set lyrics
            ..text = TextSpan(
                text: remarkLyric.lyrics,
                style: isCurrLine
                    ? currSubLyricTextStyle
                    : isDraggingLine
                        ? draggingSubLyricTextStyle
                        : subLyricTextStyle);
          //Lyrics only drawn on the screen
          if (currentLyricY < size.height && currentLyricY > 0) {
            currentSubPaint
              //Calculate text width and height
              ..layout(maxWidth: lyricMaxWidth)
              //Draw offset=Horizontal center
              ..paint(
                  canvas,
                  Offset((size.width - subLyricTextPaints[subIndex].width) / 2,
                      currentLyricY));
          }
          currentSubPaint..layout(maxWidth: lyricMaxWidth);
          //After the current lyrics are over, adjust the y coordinate of the next lyrics
          currentLyricY += currentSubPaint.height + subLyricGapValue;
        });
      }
    }
  }

  @override
  bool shouldRepaint(LyricPainter oldDelegate) {
    //Redraw when the lyrics progress changes
    return oldDelegate.currentLyricIndex != currentLyricIndex;
  }

  //Get the location of the lyrics according to the current duration
  int findLyricIndexByDuration(double curDuration, List<Lyrics> lyrics) {
    for (int i = 0; i < lyrics.length; i++) {
      if (curDuration >= lyrics[i].startTime.inMilliseconds &&
          curDuration <= lyrics[i].endTime.inMilliseconds) {
        return i;
      }
    }
    return 0;
  }

  // Calculate the offset between the incoming line and the first line
  double computeScrollY(int curLine) {
    double totalHeight = 0;
    for (var i = 0; i < curLine; i++) {
      var currPaint = lyricTextPaints[i];
      currPaint.layout(maxWidth: lyricMaxWidth);
      totalHeight += currPaint.height + lyricGapValue;
    }
    if (subLyrics != null) {
      //Increase the offset of the translated lyrics before the current line
      var list = subLyrics
          .where((subLyric) => subLyric.endTime <= lyrics[curLine].endTime)
          .toList();
      totalHeight += list.length * (subLyricGapValue + _subLyricHeight);
    }
    return totalHeight;
  }
}
